@@ -61,6 +61,11 @@ class Conf:
         # 0 unpaired electron  has S = 0,   2S + 1 = 0, called a singlet. 
         # 1 unpaired electron  has S = 1/2, 2S + 1 = 2, called a doublet (radical). 
         # 2 unpaired electrons has S = 1,   2S + 1 = 3, called a triplet.
+        self.smiles = ''
+        self.molblock = ''
+        self.symbols = []
+        self.numbers = []
+        self.positions = np.array([]) # (natoms, 3)
         self.props = {}
 
         if molecule is None:
@@ -82,10 +87,17 @@ class Conf:
                 self.rdmol = molecule
             except:
                 ValueError(f'Conf() Error: invalid Chem.Mol object')
-        self.__update()
+        
+        self._update()
 
 
-    def __update(self) -> None:
+    def _update(self) -> None:
+        self.smiles = Chem.MolToSmiles(self.rdmol)
+        self.molblock = Chem.MolToMolBlock(self.rdmol)
+        self.symbols = [a.GetSymbol() for a in self.rdmol.GetAtoms()]
+        self.numbers = [a.GetAtomicNum() for a in self.rdmol.GetAtoms()]
+        self.positions = self.rdmol.GetConformer().GetPositions() # np.ndarray
+         # check hydrogens
         num_atoms = self.rdmol.GetNumAtoms()
         tot_atoms = self.rdmol.GetNumAtoms(onlyExplicit=False)
         assert num_atoms == tot_atoms, "Conf() Error: missing hydrogens"
@@ -97,8 +109,6 @@ class Conf:
             })
         
         assert self.rdmol.GetConformer().Is3D(), "Conf() Error: not 3D"
-        
-
 
     def __str__(self) -> str:
         """Returns a string representation.
@@ -111,57 +121,7 @@ class Conf:
     ##################################################
     ### Properties
     ##################################################
-
-    @property
-    def smiles(self) -> str:
-        """Returns SMILES containing explicit hydrogens
-
-        Returns:
-            str: SMILES
-        """
-        return Chem.MolToSmiles(self.rdmol)
     
-
-    @property
-    def molblock(self) -> str:
-        """Returns MolBlock
-
-        Returns:
-            str: MolBlock
-        """
-        return Chem.MolToMolBlock(self.rdmol)
-    
-
-    @property
-    def positions(self) -> np.array:
-        """Returns the coordinates.
-
-        Returns:
-            np.array: the coordinates.
-        """
-        return np.array(self.rdmol.GetConformer().GetPositions().tolist())
-
-
-    @property
-    def symbols(self) -> list[str]:
-        """Returns the element symbols.
-
-        Returns:
-            list[str]: list of element symbols.
-        """
-        return [ a.GetSymbol() for a in self.rdmol.GetAtoms() ]
-
-
-    @property
-    def numbers(self) -> list[int]:
-        """Returns the atomic numbers.
-
-        Returns:
-            list[int]: list of atomic numbers.
-        """
-        return [ a.GetAtomicNum() for a in self.rdmol.GetAtoms() ]
-
-
     @property
     def COG(self) -> np.array:
         """Returns the center of geometry (COG).
@@ -350,7 +310,7 @@ class Conf:
             # It modifies the input molecule by adding hydrogens, 
             # but the original molecule remains unchanged.
         
-        self.__update()
+        self._update()
 
         return self
     
@@ -385,7 +345,7 @@ class Conf:
                 self.rdmol = edit_mol.GetMol()
                 Chem.SanitizeMol(self.rdmol)
         
-        self.__update()
+        self._update()
 
         return self
 
@@ -484,7 +444,7 @@ class Conf:
         
         else:
             try:
-                ase_atoms = ase.Atoms(symbols=self.symbols(), positions=self.positions())
+                ase_atoms = ase.Atoms(symbols=self.symbols, positions=self.positions)
                 ase_atoms.info['charge'] = self.charge
                 ase_atoms.info['spin'] = self.spin
                 ase_atoms.calc = calculator
@@ -501,7 +461,11 @@ class Conf:
                 raise RuntimeError("ASE calculator error")
 
 
-    def optimize(self, calculator: str | Callable = 'MMFF94', fmax: float = 0.05, max_iter: int = 1000, water: str | None = None) -> Self:
+    def optimize(self, 
+                 calculator: str | Callable = 'MMFF94', 
+                 fmax: float = 0.05, 
+                 max_iter: int = 1000, 
+                 water: str | None = None) -> Self:
         """Optimize 3D geometry.
 
         Args:
@@ -580,7 +544,7 @@ class Conf:
         else:
             # assuming ASE calculator
             with io.StringIO() as logfile:
-                ase_atoms = ase.Atoms(symbols=self.symbols(), positions=self.positions())
+                ase_atoms = ase.Atoms(symbols=self.symbols, positions=self.positions)
                 ase_atoms.info['charge'] = self.charge
                 ase_atoms.info['spin'] = self.spin
                 ase_atoms.calc = calculator
@@ -849,6 +813,253 @@ class Conf:
         return self
     
 
+    def __sp_torsion_energies(self,
+                             calculator: str | Callable,
+                             indices: tuple[int],
+                             simplify: bool = True,
+                             interval: float = 20.0,
+                             water: str | None = None) -> Self:
+        """Single-Point based potential energy profile for one torsion angle.
+
+        Args:
+            calculator (str | Callable): 'MMFF', 'UFF', 'xTB' or ASE calculator.
+            indices (tuple[int]): atom indices (i,j,k,l) for a torsion angle
+            simplify (bool, optional): whether to use fragementation. Defaults to True.
+            interval (float, optional): angle intervals. Defaults to 20.0.
+        
+        Args for xTB calculator:
+            water (str, optional): water solvation model (choose 'gbsa', 'alpb', 'cpcmx')
+                alpb: ALPB solvation model (Analytical Linearized Poisson-Boltzmann).
+                gbsa: generalized Born (GB) model with Surface Area contributions.
+                cpcmx: Extended Conductor-like Polarizable Continuum Solvation Model (CPCM-X).
+                Defaults to None.
+
+        Returns:
+            Self: modified self.
+        """
+        ref_conf = self.copy() # optimized reference conformer
+        ref_conf = ref_conf.singlepoint(calculator, water=water)
+
+        data = {'indices': indices,
+                'angle': [], 
+                'init': [], 
+                'last': [], 
+                'Converged': [],
+                }
+        
+        if simplify:
+            (frag, frag_ijkl, frag_created, wbo_filtered) = create_torsion_fragment(ref_conf.rdmol, indices)
+            frag_conf = Conf(frag)
+            for angle in np.arange(-180.0, 180.0, interval): 
+                # Iterated numpy.ndarray does not contain the last 180: -180., ..., (180).
+                conf = frag_conf.copy()
+                conf.set_torsion_angle(*frag_ijkl, angle) # atoms bonded to `l` move.
+                conf = conf.singlepoint(calculator, water=water)
+                data['angle'].append(angle)
+                # conf.singlepoint() sets `E_tot(kcal/mol)`
+                # conf.optimize() updates coordinates and conf.props: 
+                #   `E_tot_init(kcal/mol)`, `E_tot(kcal/mol)`, `Converged`.
+                data['init'].append(conf.props['E_tot(kcal/mol)'])
+                data['last'].append(ref_conf.props['E_tot(kcal/mol)'])
+                frag_cleaned, _ = clean_2d(frag, reset_isotope=True, remove_H=True)
+                # to serialize the molecule
+                data['frag'] = Chem.MolToMolBlock(frag_cleaned)
+                data['frag_indices'] = frag_ijkl
+        else:
+            for angle in np.arange(-180.0, 180.0, interval): 
+                # Iterated numpy.ndarray does not contain the last 180: -180., ..., (180).
+                conf = ref_conf.copy()
+                conf.set_torsion_angle(*indices, angle) # atoms bonded to `l` move.
+                conf = conf.singlepoint(calculator, water=water)
+                data['angle'].append(angle)
+                # conf.optimize() updates coordinates and conf.props: 
+                #   `E_tot_init(kcal/mol)`, `E_tot(kcal/mol)`, `Converged`.
+                data['init'].append(conf.props['E_tot(kcal/mol)'])
+                data['last'].append(ref_conf.props['E_tot(kcal/mol)'])
+        
+        relax = np.array(data['init']) - np.median(data['last'])
+        E_rel = relax - np.min(relax)
+
+        torsion_energy_profile = {
+            'indices' : data['indices'],
+            'angle' : np.round(np.array(data['angle']), 1).tolist(), # np.ndarray -> list for serialization
+            'E_rel(kcal/mol)': np.round(E_rel, 2).tolist(), # np.ndarray -> list for serialization
+            }
+        
+        if simplify:
+            torsion_energy_profile.update({
+                'frag' : data.get('frag', None),
+                'frag_indices' : data.get('frag_indices', None),
+            })
+
+        self.props['torsion'] = torsion_energy_profile
+
+        return self
+    
+
+    def __sp_torsion_energies_batch(self,
+                             batch_calculator: Callable,
+                             indices: tuple,
+                             simplify: bool = True,
+                             interval: float = 20.0,
+                             water: str | None = None,
+                             batchsize_atoms: int = 16*1024) -> Self:
+        """Calculate Single-Point potential energy profile for torsion angles in a batch.
+
+        Args:
+            calculator (Callable): BatchSinglePoint calculator.
+            indices (tuple): atom indices (i,j,k,l) for a torsion angle
+            simplify (bool, optional): whether to use fragementation. Defaults to True.
+            interval (float, optional): angle intervals. Defaults to 20.0.
+        
+        Args for xTB calculator:
+            water (str, optional): water solvation model (choose 'gbsa', 'alpb', 'cpcmx')
+                alpb: ALPB solvation model (Analytical Linearized Poisson-Boltzmann).
+                gbsa: generalized Born (GB) model with Surface Area contributions.
+                cpcmx: Extended Conductor-like Polarizable Continuum Solvation Model (CPCM-X).
+                Defaults to None.
+
+        Returns:
+            Self: modified self.
+        """
+        ref_conf = self.copy() # optimized reference conformer
+        evaluated_batch = batch_calculator([ref_conf.rdmol]).run()
+        for rdmol in evaluated_batch.mols:
+            ref_conf.props['E_tot(kcal/mol)'] = float(rdmol.GetProp('E_tot(kcal/mol)'))
+
+        data = {'indices': indices,
+                'angle': [], 
+                'init': [], 
+                'last': [], 
+                'Converged': [],
+                }
+        
+        confs = []
+        if simplify:
+            (frag, frag_ijkl, frag_created, wbo_filtered) = create_torsion_fragment(ref_conf.rdmol, indices)
+            frag_conf = Conf(frag)
+            for angle in np.arange(-180.0, 180.0, interval): 
+                # Iterated numpy.ndarray does not contain the last 180: -180., ..., (180).
+                conf = frag_conf.copy()
+                conf.set_torsion_angle(*frag_ijkl, angle) # atoms bonded to `l` move.
+                frag_cleaned, _ = clean_2d(frag, reset_isotope=True, remove_H=True)
+                # to serialize the molecule
+                data['frag'] = Chem.MolToMolBlock(frag_cleaned)
+                data['frag_indices'] = frag_ijkl
+                data['angle'].append(angle)
+                confs.append(conf)
+      
+        else:
+            for angle in np.arange(-180.0, 180.0, interval): 
+                # Iterated numpy.ndarray does not contain the last 180: -180., ..., (180).
+                conf = ref_conf.copy()
+                conf.set_torsion_angle(*indices, angle) # atoms bonded to `l` move.
+                data['angle'].append(angle)
+                confs.append(conf)
+                
+        batches = prepare_batches(confs, batchsize_atoms)
+
+        for batch in batches:
+            evaluated_batch = batch_calculator(batch.rdmols).run()
+            for rdmol in evaluated_batch.mols:
+                # batch single point calculator sets 'E_tot(kcal/mol)'
+                # batch optimizer updates coordinates and conf.props: 
+                #   `E_tot_init(kcal/mol)`, `E_tot(kcal/mol)`, `Converged`.
+                data['init'].append(float(rdmol.GetProp('E_tot(kcal/mol)')))
+                data['last'].append(ref_conf.props['E_tot(kcal/mol)'])
+        
+        relax = np.array(data['init']) - np.median(data['last'])
+        E_rel = relax - np.min(relax)
+
+        torsion_energy_profile = {
+            'indices' : data['indices'],
+            'angle' : np.round(np.array(data['angle']), 1).tolist(), # np.ndarray -> list for serialization
+            'E_rel(kcal/mol)': np.round(E_rel, 2).tolist(), # np.ndarray -> list for serialization
+            }
+        
+        if simplify:
+            torsion_energy_profile.update({
+                'frag' : data.get('frag', None),
+                'frag_indices' : data.get('frag_indices', None),
+            })
+
+        self.props['torsion'] = torsion_energy_profile
+
+        return self
+    
+
+    def calculate_sp_torsion_energies(self,
+                         calculator: str | Callable,
+                         torsion_angle_idx: int | None = None,
+                         simplify: bool = True, 
+                         interval: float = 20.0,
+                         water: str | None = None,
+                         batchsize_atoms: int = 0,
+                         ) -> Self:
+        """Calculates Single-Point potential energy profiles for each torsion angle using ASE or BatchSinglePoint.
+
+        Args:
+            calculator (str | Callable): 'MMFF', 'UFF', 'xTB', ASE calculator, or BatchOptimizer.
+            torsion_angle_idx (int | None): key to the torsion indices to calculate. Defaults to None (all).
+            simplify (bool, optional): whether to use fragment surrogate. Defaults to True.
+            fmax (float, optional): fmax of ASE optimizer. Defaults to 0.05.
+            interval (float, optional): interval of torsion angles in degree. Defaults to 15.0.
+            use_converged_only (bool, optional): whether to use only converged data. Defaults to True.
+            batchsize_atoms (int, optional): maximum number of atoms in a single batch.
+                Setting any number smaller than conf.natoms to disable batch optimization.
+                Defaults to 0.
+        
+        Args for xTB calculator:
+            water (str, optional): water solvation model (choose 'gbsa' or 'alpb')
+                alpb: ALPB solvation model (Analytical Linearized Poisson-Boltzmann).
+                gbsa: generalized Born (GB) model with Surface Area contributions.
+                cpcmx: Extended Conductor-like Polarizable Continuum Solvation Model (CPCM-X).
+                Defaults to None.
+
+        Returns:
+            Self: modified self.
+        """
+
+        torsion_angle_atom_indices_list = self.get_torsion_angle_atoms() 
+        # [(5, 4, 3, 1), ...]
+
+        if isinstance(torsion_angle_idx, int) and torsion_angle_idx < len(torsion_angle_atom_indices_list):
+            indices = {torsion_angle_idx: torsion_angle_atom_indices_list[torsion_angle_idx]}
+            # atom indices for a single torsion angle
+            # ex. {0: (5,4,3,1)}
+        else:
+            indices = {k:v for k, v in enumerate(torsion_angle_atom_indices_list)}
+            # atom indices for all torsion angles
+            # ex. {0: (5,4,3,1), 1: ...}
+
+        conf = self.copy()
+        torsion_energies = []
+        if batchsize_atoms >= conf.natoms:
+            for torsion_angle_idx, ijkl in indices.items():
+                conf = conf.__sp_torsion_energies_batch(calculator, 
+                                                        ijkl, 
+                                                        simplify, 
+                                                        interval,
+                                                        water=water,
+                                                        batchsize_atoms=batchsize_atoms,
+                                                        )
+                conf.props['torsion'].update({'torsion_angle_idx': torsion_angle_idx})
+                torsion_energies.append(conf.props['torsion'])
+        else:
+            for torsion_angle_idx, ijkl in indices.items():
+                conf = conf.__sp_torsion_energies(calculator, 
+                                                  ijkl, 
+                                                  simplify, 
+                                                  interval, 
+                                                  water)
+                conf.props['torsion'].update({'torsion_angle_idx': torsion_angle_idx})
+                torsion_energies.append(conf.props['torsion'])
+        
+        self.props['torsion'] = torsion_energies
+
+        return self
+    
+
     def dumps(self, key: str = '') -> str:
         """Returns JSON dumps of the `props`.
 
@@ -906,10 +1117,33 @@ class Conf:
         self.spin = int(data['spin'])
         self.props = data['props']
         self.rdmol = Chem.MolFromMolBlock(data['molblock'], sanitize=False, removeHs=False)
+        self._update()
 
         return self
 
 
+    def to_geometry(self) -> str:
+        """Returns geometry input for psi4 or other quantum chemistry software.
+
+        Each line has atom symbol and its X, Y, and Z coordinates.
+        Unit is Angstrom.
+
+        O  0.0  0.0  0.0
+        H  0.757  0.586  0.0
+        H -0.757  0.586  0.0
+        
+        Example:
+            import psi4
+            geometry = conf.to_geometry()
+            mol = psi4.geometry(geometry)
+
+        Returns:
+            str: geometry input for psi4 or other quantum chemistry software.
+        """
+        lines = [f'{e:5} {x:.3f} {y:.3f} {z:.3f}' for e, (x, y, z) in zip(self.symbols, self.positions)]
+        return '\n'.join(lines)
+    
+
     def to_xyz(self) -> str:
         """Returns XYZ formatted strings.
 
@@ -917,7 +1151,7 @@ class Conf:
             str: XYZ formatted strings.
         """
         lines = [f'{self.natoms}', ' ']
-        for e, (x, y, z) in zip(self.symbols(), self.positions()):
+        for e, (x, y, z) in zip(self.symbols, self.positions):
             lines.append(f'{e:5} {x:23.14f} {y:23.14f} {z:23.14f}')
         return '\n'.join(lines)
 
@@ -950,7 +1184,7 @@ class Conf:
         else:
             lines = ["$coord angs"]
 
-        for (x, y, z), e in zip(self.positions(), self.symbols()):
+        for (x, y, z), e in zip(self.positions, self.symbols):
             lines.append(f"{x:20.15f} {y:20.15f} {z:20.15f} {e}")
         
         lines.append("$end")
