@@ -10,6 +10,8 @@ from pathlib import Path
 from collections import defaultdict, namedtuple
 from collections.abc import Callable
 from typing import Iterator, Self
+from concurrent.futures import ProcessPoolExecutor
+from tqdm import tqdm
 
 import numpy as np
 # import pandas as pd
@@ -486,7 +488,12 @@ class Mol:
         
         return self
 
+
+    def count(self) -> int:
+        """Returns the number of conformers"""
+        return len(self.confs)
     
+
     def copy(self) -> Self:
         """Returns a copy of self.
 
@@ -730,12 +737,26 @@ class Mol:
         return self
     
 
+    @staticmethod
+    def _map_optimize(conf:Conf, targs:tuple) -> Conf:
+        """A mask function to return True if molecule is similar with target molecules, `targs`.
+
+        Args:
+            mol (Mol): subject rdworks.Mol object.
+            targs (tuple): a tuple of rdworks.Mol objects to compare.
+
+        Returns:
+            bool: True if molecule is similar with target molecules.
+        """
+        return conf.optimize(*targs) # unpack tuple of arguments
+
+
     def optimize_confs(self, 
                        calculator: str | Callable = 'MMFF94', 
                        fmax: float = 0.05, 
                        max_iter: int = 1000,
+                       water: str | None = None,
                        batchsize_atoms: int = 0,
-                       water: str | None = None
                        ) -> Self:
         """Optimizes 3D geometry of conformers.
 
@@ -768,10 +789,22 @@ class Mol:
         Returns:
             Self: modified self.
         """
-        if isinstance(calculator, Callable) and batchsize_atoms >= self.confs[0].natoms:
-            self.confs = batch_optimize(self.confs, calculator, batchsize_atoms)
+        if isinstance(calculator, Callable):
+            if batchsize_atoms >= self.confs[0].natoms:
+                self.confs = batch_optimize(self.confs, calculator, batchsize_atoms)
+            else:
+                self.confs = [conf.optimize(calculator, fmax, max_iter, water) for conf in self.confs]
         else:
-            self.confs = [ conf.optimize(calculator, fmax, max_iter, water=water) for conf in self.confs ]
+            if self.max_workers > 1:
+                largs = [(calculator, fmax, max_iter, water) for i in self.confs]
+                with ProcessPoolExecutor(max_workers=self.max_workers) as executor:
+                    if self.progress:
+                        self.confs = list(tqdm(executor.map(Mol._map_optimize, self.confs, largs, chunksize=1),
+                            desc="Optimize Conformers", total=self.count()))
+                    else:
+                        self.confs = list(executor.map(Mol._map_optimize, self.confs, largs, chunksize=1))
+            else:
+                self.confs = [conf.optimize(calculator, fmax, max_iter, water) for conf in self.confs]
 
         return self
 
@@ -1282,6 +1315,56 @@ class Mol:
         return self
     
 
+    def calculate_sp_torsion_energies(self,
+                         calculator: str | Callable = 'MMFF94',
+                         torsion_angle_idx: int | None = None,
+                         simplify: bool = True, 
+                         interval: float = 20.0,
+                         water: str | None = None,
+                         batchsize_atoms: int = 0,
+                         ) -> Self:
+        """Calculates potential energy profiles for each torsion angle using ASE optimizer.
+
+        It uses the first conformer as a reference.
+
+        Args:
+            calculator (str | Callable): 'MMFF', 'UFF', 'xTB' or ASE calculator.
+            torsion_angle_idx (int | None): torsion index to calculate. Defaults to None (all).
+            simplify (bool, optional): whether to use fragment surrogate. Defaults to True.
+            interval (float, optional): interval of torsion angles in degree. Defaults to 15.0.
+            batchsize_atoms (int, optional): maximum number of atoms in a single batch.
+                Setting any number smaller than conf.natoms to disable batch optimization.
+                Defaults to 0.
+        
+        Args for xTB calculator:
+            water (str, optional): water solvation model (choose 'gbsa', 'alpb', or 'cpcmx')
+                alpb: ALPB solvation model (Analytical Linearized Poisson-Boltzmann).
+                gbsa: generalized Born (GB) model with Surface Area contributions.
+                cpcmx: Extended Conductor-like Polarizable Continuum Solvation Model (CPCM-X).
+                Defaults to None.
+
+        Returns:
+            Self: modified self.
+        """
+        assert self.num_confs > 0, "calculate_sp_torsion_energies() requires at least one conformer"
+        
+        ref_conf = self.confs[0].copy()
+        
+        if batchsize_atoms >= ref_conf.natoms:
+            assert not isinstance(calculator, str), "Batch singlepoint is required for batch calculation"
+
+        ref_conf = ref_conf.calculate_sp_torsion_energies(calculator, 
+                                            torsion_angle_idx, 
+                                            simplify, 
+                                            interval, 
+                                            water=water,
+                                            batchsize_atoms=batchsize_atoms)
+        
+        self.props['torsion'] = ref_conf.props['torsion']
+
+        return self
+    
+
     def draw(self, 
              coordgen: bool = False,
              rotate: bool = False,
