@@ -467,9 +467,9 @@ def create_torsion_fragment(rdmol: Chem.Mol,
         # no fragmentation
         return (rdmol, torsion_indices, False, False)
 
-    # fragmented
-    WBO_filtered = False
-    if GFN2xTB.is_ready():
+    if GFN2xTB().is_ready():
+        # fragmented
+        WBO_filtered = False
         # filter candidate(s) by Wiberg bond order (WBO) if xTB is available
         jk = tuple(sorted([j, k]))
         wbo_passed_candidates = {}
