@@ -7,6 +7,7 @@ import logging
 import shutil
 import re
 import numpy as np
+
 from scipy.spatial import cKDTree
 from scipy.interpolate import griddata
 
@@ -32,173 +33,25 @@ ev2kcalpermol = 23.060547830619026
 pt = Chem.GetPeriodicTable()
 
 
-def read_xyz_file(filename):
-    """Read molecular structure from XYZ file."""
-    with open(filename, 'r') as f:
-        lines = f.readlines()
-    
-    natoms = int(lines[0].strip())
-    atoms = []
-
-    for i in range(2, 2 + natoms):
-        parts = lines[i].split()
-        element = parts[0]
-        x, y, z = float(parts[1]), float(parts[2]), float(parts[3])
-        try:
-            atomic_num = pt.GetAtomicNumber(element)
-        except:
-            atomic_num = 6
-        atoms.append([atomic_num, x, y, z])
-    
-    return atoms
-
-
-def read_surface_esp(filename):
-    """
-    Read xtb ESP surface file with format: x y z esp_value
-    """
-    points = []
-    with open(filename, 'r') as f:
-        for line in f:
-            parts = line.split()
-            if len(parts) >= 4:
-                try:
-                    x = float(parts[0])
-                    y = float(parts[1])
-                    z = float(parts[2])
-                    esp = float(parts[3])
-                    points.append([x, y, z, esp])
-                except ValueError:
-                    continue
-    
-    return np.array(points)
-
-
-def create_regular_grid(surface_points, atoms, grid_spacing=0.6, padding=4.0):
-    """
-    Create a regular 3D grid encompassing the molecule.
-    
-    Args:
-        surface_points: Nx4 array of surface points (x, y, z, esp)
-        atoms: List of atomic coordinates
-        grid_spacing: Spacing between grid points (in Bohr or Angstrom)
-        padding: Extra space around molecule
-    """
-    # Get molecule bounds
-    all_coords = np.array([[a[1], a[2], a[3]] for a in atoms])
-    
-    # Determine grid bounds with padding
-    min_coords = np.min(all_coords, axis=0) - padding
-    max_coords = np.max(all_coords, axis=0) + padding
-    
-    # Also consider surface points
-    surface_coords = surface_points[:, :3]
-    min_coords = np.minimum(min_coords, np.min(surface_coords, axis=0) - padding)
-    max_coords = np.maximum(max_coords, np.max(surface_coords, axis=0) + padding)
-    
-    # Create grid
-    x_range = np.arange(min_coords[0], max_coords[0] + grid_spacing, grid_spacing)
-    y_range = np.arange(min_coords[1], max_coords[1] + grid_spacing, grid_spacing)
-    z_range = np.arange(min_coords[2], max_coords[2] + grid_spacing, grid_spacing)
-    
-    nx, ny, nz = len(x_range), len(y_range), len(z_range)
-    
-    logger.info(f"Grid dimensions: {nx} x {ny} x {nz} = {nx*ny*nz} points")
-    logger.info(f"Grid spacing: {grid_spacing:.3f}")
-    logger.info(
-        f"Grid bounds: X[{min_coords[0]:.2f}, {max_coords[0]:.2f}], "
-        f"Y[{min_coords[1]:.2f}, {max_coords[1]:.2f}], "
-        f"Z[{min_coords[2]:.2f}, {max_coords[2]:.2f}]")
-    
-    # Create 3D meshgrid
-    X, Y, Z = np.meshgrid(x_range, y_range, z_range, indexing='ij')
-    grid_points = np.c_[X.ravel(), Y.ravel(), Z.ravel()]
-    
-    origin = [x_range[0], y_range[0], z_range[0]]
-    grid_vectors = [
-        [grid_spacing, 0.0, 0.0],
-        [0.0, grid_spacing, 0.0],
-        [0.0, 0.0, grid_spacing]
-    ]
-    
-    return grid_points, (nx, ny, nz), origin, grid_vectors
-
-
-def interpolate_esp_to_grid(surface_points, grid_points, method='nearest', max_distance=2.0):
-    """
-    Interpolate surface ESP values to regular grid using nearest neighbor.
-    Points far from surface are set to zero.
-    
-    Args:
-        surface_points: Nx4 array (x, y, z, esp)
-        grid_points: Mx3 array of grid coordinates
-        method: Interpolation method ('nearest', 'linear', or 'cubic')
-        max_distance: Maximum distance for interpolation (points further get 0)
-    """
-    surface_coords = surface_points[:, :3]
-    esp_values = surface_points[:, 3]
-    
-    logger.info(f"Interpolating {len(grid_points)} grid points from {len(surface_points)} surface points...")
-    
-    # Build KD-tree for efficient nearest neighbor search
-    tree = cKDTree(surface_coords)
-    
-    # Find distances to nearest surface point
-    distances, indices = tree.query(grid_points, k=1)
-    
-    # Interpolate using scipy
-    if method == 'nearest':
-        grid_esp = esp_values[indices]
-    else:
-        grid_esp = griddata(surface_coords, esp_values, grid_points, method=method, fill_value=0.0)
-    
-    # Zero out points far from surface
-    grid_esp[distances > max_distance] = 0.0
-    
-    logger.info(f"ESP range on grid: {np.min(grid_esp):.4f} to {np.max(grid_esp):.4f}")
-    logger.info(f"Points within {max_distance:.1f} of surface: {np.sum(distances <= max_distance)}")
-    
-    return grid_esp
-
-
-def write_cube_file(filename, atoms, grid_dims, origin, grid_vectors, values):
-    """Write Gaussian cube file format."""
-    nx, ny, nz = grid_dims
-    
-    with open(filename, 'w') as f:
-        f.write("Cube file: ESP interpolated from xtb surface data\n")
-        f.write("Electrostatic potential (a.u.)\n")
-        
-        # Number of atoms and origin
-        f.write(f"{len(atoms):5d} {origin[0]:12.6f} {origin[1]:12.6f} {origin[2]:12.6f}\n")
-        
-        # Grid dimensions and vectors
-        f.write(f"{nx:5d} {grid_vectors[0][0]:12.6f} {grid_vectors[0][1]:12.6f} {grid_vectors[0][2]:12.6f}\n")
-        f.write(f"{ny:5d} {grid_vectors[1][0]:12.6f} {grid_vectors[1][1]:12.6f} {grid_vectors[1][2]:12.6f}\n")
-        f.write(f"{nz:5d} {grid_vectors[2][0]:12.6f} {grid_vectors[2][1]:12.6f} {grid_vectors[2][2]:12.6f}\n")
-        
-        # Atom coordinates
-        for atom in atoms:
-            atomic_num, x, y, z = atom
-            f.write(f"{atomic_num:5d} {float(atomic_num):12.6f} {x:12.6f} {y:12.6f} {z:12.6f}\n")
-        
-        # ESP values (6 per line)
-        for i in range(0, len(values), 6):
-            line_values = values[i:i+6]
-            f.write(" ".join([f"{v:13.5E}" for v in line_values]) + "\n")
-
-
 class GFN2xTB:
-    def __init__(self, molecule: Chem.Mol, ncores: int | None = None):
-        assert isinstance(molecule, Chem.Mol), "molecule is not rdkit.Chem.Mol type"
-        assert molecule.GetConformer().Is3D(), "molecule is not a 3D conformer"
-        assert self.is_xtb_ready(), "xtb is not accessible"
+    def __init__(self, molecule: Chem.Mol | None = None, ncores: int | None = None, xtb_exec: str | Path | None = None):
+        """_summary_
 
-        self.rdmol = molecule
-        self.charge = rdmolops.GetFormalCharge(self.rdmol)
-        self.natoms = molecule.GetNumAtoms()
-        self.symbols = [ atom.GetSymbol() for atom in molecule.GetAtoms() ]
-        self.positions = molecule.GetConformer().GetPositions().tolist()
+        Args:
+            molecule (Chem.Mol): input molecule in rdkit.Chem.Mol format
+            ncores (int | None, optional): number of cores for parallelization. Defaults to None (all).
+            xtb_exec (str | Path | None, optional): path to xtb executable. Defaults to None.
+        """
+        if molecule is not None:
+            assert isinstance(molecule, Chem.Mol), "molecule must be rdkit.Chem.Mol type"
+            assert molecule.GetConformer().Is3D(), "molecule must be a 3D conformer"
+
+            self.rdmol = molecule
+            self.charge = rdmolops.GetFormalCharge(self.rdmol)
+            self.natoms = molecule.GetNumAtoms()
+            self.symbols = [ atom.GetSymbol() for atom in molecule.GetAtoms() ]
+            self.numbers = [ atom.GetAtomicNum() for atom in molecule.GetAtoms() ]
+            self.positions = molecule.GetConformer().GetPositions().tolist()
 
         if ncores is None:
             ncores = os.cpu_count()
@@ -215,19 +68,27 @@ class GFN2xTB:
         # unlimit the system stack
         resource.setrlimit(resource.RLIMIT_STACK, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))
 
+        self.xtb_exec = None
+        self.cpx_exec = None
 
-    @staticmethod
-    def is_xtb_ready() -> bool:
+        if xtb_exec is None:
+            self.xtb_exec = shutil.which('xtb')
+            self.cpx_exec = shutil.which('cpx')
+        else:
+            self.xtb_exec = Path(xtb_exec).resolve().as_posix()
+            self.cpx_exec = Path(xtb_exec).resolve().parent.joinpath('cpx').as_posix()
+
+
+    def is_xtb_ready(self) -> bool:
         """Check if xtb is available.
 
         Returns:
             bool: True if `xtb` is available, False otherwise.
         """
-        return shutil.which('xtb') is not None
+        return self.xtb_exec is not None
 
 
-    @staticmethod
-    def is_optimize_ready() -> bool:
+    def is_optimize_ready(self) -> bool:
         try:
             h2o = [
                 '$coord',
@@ -241,7 +102,7 @@ class GFN2xTB:
                 test_geometry = os.path.join(temp_dir, 'coord')
                 with open(test_geometry, 'w') as f:
                     f.write('\n'.join(h2o))
-                proc = subprocess.run(['xtb', test_geometry, '--opt'],
+                proc = subprocess.run([self.xtb_exec, test_geometry, '--opt'],
                                       cwd=temp_dir,
                                       capture_output=True, 
                                       text=True,
@@ -265,18 +126,16 @@ $ cp -r xtb-dist/share      /usr/local/ """)
             return False
 
 
-    @staticmethod
-    def is_cpx_ready() -> bool:
+    def is_cpx_ready(self) -> bool:
         """Checks if the CPCM-X command-line tool, `cpx`, is accessible in the system.
 
         Returns:
             bool: True if the cpx is found, False otherwise.
         """
-        return shutil.which('cpx') is not None
+        return self.cpx_exec is not None
     
 
-    @staticmethod
-    def is_cpcmx_ready() -> bool:
+    def is_cpcmx_ready(self) -> bool:
         """Checks if xtb works with the `--cpcmx` option.
 
         xtb distributed by the conda does not include CPCM-X function (as of June 17, 2025). 
@@ -285,9 +144,9 @@ $ cp -r xtb-dist/share      /usr/local/ """)
         Returns:
             bool: True if the --cpcmx option is working, False otherwise.
         """
-        if GFN2xTB.is_xtb_ready():
+        if self.is_xtb_ready():
             with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
-                cmd = ['xtb', '--cpcmx']
+                cmd = [self.xtb_exec, '--cpcmx']
                 proc = subprocess.run(cmd, cwd=temp_dir, capture_output=True, text=True, encoding='utf-8')
                 # we are expecting an error because no input file is given
                 assert proc.returncode != 0
@@ -299,29 +158,27 @@ $ cp -r xtb-dist/share      /usr/local/ """)
         return True
 
 
-    @staticmethod
-    def is_ready() -> bool:
+    def is_ready(self) -> bool:
         """Check if `xtb` and `cpx` are accessible and `xtb --cpcmx` are available.
 
         Returns:
             bool: True if both `xtb` and `cpx` are accessible, False otherwise.
         """
-        return all([GFN2xTB.is_xtb_ready(),
-                    GFN2xTB.is_cpx_ready(),
-                    GFN2xTB.is_cpcmx_ready(),
-                    GFN2xTB.is_optimize_ready()])
+        return all([self.is_xtb_ready(),
+                    self.is_cpx_ready(),
+                    self.is_cpcmx_ready(),
+                    self.is_optimize_ready()])
     
 
-    @staticmethod
-    def version() -> str | None:
+    def version(self) -> str | None:
         """Check xtb version.
 
         Returns:
             str | None: version statement.
         """
-        if GFN2xTB.is_xtb_ready():
+        if self.is_xtb_ready():
             with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
-                cmd = ['xtb', '--version']
+                cmd = [self.xtb_exec, '--version']
                 proc = subprocess.run(cmd, cwd=temp_dir, capture_output=True, text=True, encoding='utf-8')
                 assert proc.returncode == 0, "GFN2xTB() Error: xtb not available"
                 match = re.search('xtb\s+version\s+(?P<version>[\d.]+)', proc.stdout)
@@ -506,7 +363,7 @@ $ cp -r xtb-dist/share      /usr/local/ """)
             with open(geometry_input_path, 'w') as geometry:
                 geometry.write(self.to_xyz())
             
-            cmd = ['xtb', geometry_input_path.as_posix()]
+            cmd = [self.xtb_exec, geometry_input_path.as_posix()]
             options = ['-c', str(self.charge), '--sp', '--gfn', '2', '--json']
             
             if water is not None and isinstance(water, str):
@@ -613,7 +470,7 @@ $ cp -r xtb-dist/share      /usr/local/ """)
             with open(geometry_input_path, 'w') as geometry:
                 geometry.write(self.to_xyz())
 
-            cmd = ['xtb', geometry_input_path.as_posix()]
+            cmd = [self.xtb_exec, geometry_input_path.as_posix()]
             options = ['-c', str(self.charge), '-o', 'normal', '--gfn', '2', '--json']
 
             if water is not None and isinstance(water, str):
@@ -650,7 +507,144 @@ $ cp -r xtb-dist/share      /usr/local/ """)
         return SimpleNamespace()
 
 
-    
+    @staticmethod
+    def read_esp_data(filename: str | Path) -> np.ndarray:
+        """
+        Read xtb ESP surface file with format: x y z esp_value
+        """
+        points = []
+        with open(filename, 'r') as f:
+            for line in f:
+                parts = line.strip().split()
+                if len(parts) >= 4:
+                    try:
+                        x = float(parts[0])
+                        y = float(parts[1])
+                        z = float(parts[2])
+                        esp = float(parts[3])
+                        points.append((x, y, z, esp))
+                    except ValueError:
+                        continue
+        
+        return np.array(points)
+
+
+    def create_regular_grid(self, surface_points, grid_spacing=0.6, padding=4.0, verbose: bool = False):
+        """
+        Create a regular 3D grid encompassing the molecule.
+        
+        Args:
+            surface_points: Nx4 array of surface points (x, y, z, esp)
+            atoms: List of atomic coordinates
+            grid_spacing: Spacing between grid points (in Bohr or Angstrom)
+            padding: Extra space around molecule
+        """
+        # Get molecule bounds
+        all_coords = np.array(self.positions)
+        
+        # Determine grid bounds with padding
+        min_coords = np.min(all_coords, axis=0) - padding
+        max_coords = np.max(all_coords, axis=0) + padding
+        
+        # Also consider surface points
+        surface_coords = surface_points[:, :3]
+        min_coords = np.minimum(min_coords, np.min(surface_coords, axis=0) - padding)
+        max_coords = np.maximum(max_coords, np.max(surface_coords, axis=0) + padding)
+        
+        # Create grid
+        x_range = np.arange(min_coords[0], max_coords[0] + grid_spacing, grid_spacing)
+        y_range = np.arange(min_coords[1], max_coords[1] + grid_spacing, grid_spacing)
+        z_range = np.arange(min_coords[2], max_coords[2] + grid_spacing, grid_spacing)
+        
+        nx, ny, nz = len(x_range), len(y_range), len(z_range)
+        
+        if verbose:
+            logger.info(f"Grid dimensions: {nx} x {ny} x {nz} = {nx*ny*nz} points")
+            logger.info(f"Grid spacing: {grid_spacing:.3f}")
+            logger.info(
+                f"Grid bounds: X[{min_coords[0]:.2f}, {max_coords[0]:.2f}], "
+                f"Y[{min_coords[1]:.2f}, {max_coords[1]:.2f}], "
+                f"Z[{min_coords[2]:.2f}, {max_coords[2]:.2f}]")
+        
+        # Create 3D meshgrid
+        X, Y, Z = np.meshgrid(x_range, y_range, z_range, indexing='ij')
+        grid_points = np.c_[X.ravel(), Y.ravel(), Z.ravel()]
+        
+        origin = [x_range[0], y_range[0], z_range[0]]
+        grid_vectors = [
+            [grid_spacing, 0.0, 0.0],
+            [0.0, grid_spacing, 0.0],
+            [0.0, 0.0, grid_spacing]
+        ]
+        
+        return grid_points, (nx, ny, nz), origin, grid_vectors
+
+
+    @staticmethod
+    def interpolate_esp_to_grid(surface_points, grid_points, method='nearest', max_distance=2.0):
+        """
+        Interpolate surface ESP values to regular grid using nearest neighbor.
+        Points far from surface are set to zero.
+        
+        Args:
+            surface_points: Nx4 array (x, y, z, esp)
+            grid_points: Mx3 array of grid coordinates
+            method: Interpolation method ('nearest', 'linear', or 'cubic')
+            max_distance: Maximum distance for interpolation (points further get 0)
+        """
+        surface_coords = surface_points[:, :3]
+        esp_values = surface_points[:, 3]
+        
+        logger.info(f"Interpolating {len(grid_points)} grid points from {len(surface_points)} surface points...")
+        
+        # Build KD-tree for efficient nearest neighbor search
+        tree = cKDTree(surface_coords)
+        
+        # Find distances to nearest surface point
+        distances, indices = tree.query(grid_points, k=1)
+        
+        # Interpolate using scipy
+        if method == 'nearest':
+            grid_esp = esp_values[indices]
+        else:
+            grid_esp = griddata(surface_coords, esp_values, grid_points, method=method, fill_value=0.0)
+        
+        # Zero out points far from surface
+        grid_esp[distances > max_distance] = 0.0
+        
+        logger.info(f"ESP range on grid: {np.min(grid_esp):.4f} to {np.max(grid_esp):.4f}")
+        logger.info(f"Points within {max_distance:.1f} of surface: {np.sum(distances <= max_distance)}")
+        
+        return grid_esp
+
+
+    def to_cube(self, grid_dims, origin, grid_vectors, values) -> str:
+        """Write Gaussian cube file format."""
+        nx, ny, nz = grid_dims
+        cube = []
+        cube.append("Cube file: ESP interpolated from xtb surface data")
+        cube.append("Electrostatic potential (a.u.)")
+            
+        # Number of atoms and origin
+        cube.append(f"{self.natoms:5d} {origin[0]:12.6f} {origin[1]:12.6f} {origin[2]:12.6f}")
+        
+        # Grid dimensions and vectors
+        cube.append(f"{nx:5d} {grid_vectors[0][0]:12.6f} {grid_vectors[0][1]:12.6f} {grid_vectors[0][2]:12.6f}")
+        cube.append(f"{ny:5d} {grid_vectors[1][0]:12.6f} {grid_vectors[1][1]:12.6f} {grid_vectors[1][2]:12.6f}")
+        cube.append(f"{nz:5d} {grid_vectors[2][0]:12.6f} {grid_vectors[2][1]:12.6f} {grid_vectors[2][2]:12.6f}")
+            
+        # Atom coordinates
+        for atomic_num, (x, y, z) in zip(self.numbers, self.positions):
+            cube.append(f"{atomic_num:5d} {float(atomic_num):12.6f} {x:12.6f} {y:12.6f} {z:12.6f}")
+        
+        # ESP values (6 per line)
+        for i in range(0, len(values), 6):
+            line_values = values[i:i+6]
+            cube.append(" ".join([f"{v:13.5E}" for v in line_values]))
+
+        return '\n'.join(cube)
+
+
     def esp_surface_points(self, water: str | None = None, verbose: bool = False) -> str:
         """Calculate electrostatic potential surface points data.
         
@@ -690,7 +684,7 @@ $ cp -r xtb-dist/share      /usr/local/ """)
             with open(geometry_input_path, 'w') as geometry:
                 geometry.write(self.to_xyz())
 
-            cmd = ['xtb', geometry_input_path.as_posix()]
+            cmd = [self.xtb_exec, geometry_input_path.as_posix()]
 
             options = ['--sp', '--esp', '--gfn', '2']
 
@@ -711,12 +705,13 @@ $ cp -r xtb-dist/share      /usr/local/ """)
 
 
     def esp_volumetric(self, 
-            water: str | None = None, 
+            water: str | None = None,
+            max_iterations: int = 250,
             grid_spacing: float = 0.6,
             padding: float = 4.0,
             max_distance: float = 2.0,
             method: str = 'nearest',
-            verbose: bool = False) -> tuple[str, float, float]:
+            verbose: bool = False) -> str:
         """Calculate electrostatic potential volumetric data.
         
         Example:
@@ -730,60 +725,48 @@ $ cp -r xtb-dist/share      /usr/local/ """)
             view.zoomTo()
             view.show()
         """
-        esp_voldata = ""
-        esp_scale_min = 0.0
-        esp_scale_max = 0.0
         with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
             workdir = Path(temp_dir)
             if verbose:
                 logger.info(f'xtb.optimize workdir= {temp_dir}')
             
-            geometry_input_path = workdir / 'geometry.xyz'
-            xtb_esp_dat = workdir / 'xtb_esp.dat'
-            esp_cube = workdir / 'esp.cube'
-
+            geometry_input_path = workdir / 'geometry.xyz' # input
+            
             with open(geometry_input_path, 'w') as geometry:
                 geometry.write(self.to_xyz())
 
-            cmd = ['xtb', geometry_input_path.as_posix()]
+            cmd = [self.xtb_exec, geometry_input_path.as_posix()]
 
-            options = ['--sp', '--esp', '--gfn', '2']
+            options = ['--sp', 
+                       '--esp', 
+                       '--gfn', '2', 
+                       '--iterations', str(max_iterations),
+                       ]
 
             if water is not None and isinstance(water, str):
                 if water == 'gbsa':
                     options += ['--gbsa', 'H2O']
                 elif water == 'alpb':
                     options += ['--alpb', 'water']
+                elif water == 'cpcmx' and self.is_cpcmx_ready():
+                    options += ['--cpcmx', 'water']
 
             proc = subprocess.run(cmd + options, cwd=temp_dir, capture_output=True, text=True, encoding='utf-8')
-            # output files: xtb_esp.cosmo, xtb_esp.dat, xtb_esp_profile.dat
-
-            if proc.returncode == 0 and xtb_esp_dat.is_file():                
-                # [1/5] Reading molecular structure
-                atoms = read_xyz_file(geometry_input_path)                
-                
-                # [2/5] Reading surface ESP data
-                surface_points = read_surface_esp(xtb_esp_dat)
+            
+            try:
+                # output files: xtb_esp.cosmo, xtb_esp.dat, xtb_esp_profile.dat
+                xtb_esp_dat = workdir / 'xtb_esp.dat' # expected output
+                assert proc.returncode == 0
+                assert xtb_esp_dat.is_file()
+                surface_points = GFN2xTB.read_esp_data(xtb_esp_dat)
                 if verbose:
                     logger.info(f"Loaded {len(surface_points)} surface points")
                     logger.info(f"ESP range: {np.min(surface_points[:, 3]):.4f} to {np.max(surface_points[:, 3]):.4f}")
-                
-                # [3/5] Creating regular grid
-                grid_points, grid_dims, origin, grid_vectors = create_regular_grid(surface_points, atoms, grid_spacing, padding)
-                
-                # [4/5] Interpolating ESP to grid
-                grid_esp = interpolate_esp_to_grid(surface_points, grid_points, method, max_distance)
-                
-                # [5/5] Writing cube file
-                write_cube_file(esp_cube, atoms, grid_dims, origin, grid_vectors, grid_esp)
-                esp_scale_min = np.min(grid_esp)
-                esp_scale_max = np.max(grid_esp)
-                
+                grid_points, grid_dims, origin, grid_vectors = self.create_regular_grid(surface_points, grid_spacing, padding)
                 if verbose:
                     logger.info(f"Grid: {grid_dims[0]} x {grid_dims[1]} x {grid_dims[2]} points")
-                    logger.info(f"ESP range: {esp_scale_min:.4f} to {esp_scale_max:.4f} a.u.")
-                    
-                with open(esp_cube, 'r') as f:
-                    esp_voldata = f.read()
-    
-        return (esp_voldata, esp_scale_min, esp_scale_max)
\ No newline at end of file
+                grid_esp = GFN2xTB.interpolate_esp_to_grid(surface_points, grid_points, method, max_distance)
+                return self.to_cube(grid_dims, origin, grid_vectors, grid_esp)
+            except Exception as e:
+                logger.error(f"Failed to generate ESP volumetric data ({e}).")
+                return ""
\ No newline at end of file
