@@ -8,14 +8,25 @@ from rdworks import State, StateEnsemble, StateNetwork
 
 
 @pytest.fixture(scope='module') # Runs once for every test module (file).
-def prepared_state():
-    smiles = 'c1ccc(CNc2ncnc3ccccc23)cc1'
+def prepared_sm07_state():
+    smiles = 'c1ccc(CNc2ncnc3ccccc23)cc1' # SAMPL6 challenge SM07
     st = State(smiles=smiles)
     return st
 
 
-def test_state_serialize(prepared_state):
-    st = prepared_state
+@pytest.fixture(scope='module')
+def prepared_sm07_state_ensemble():
+    smiles = 'c1ccc(CNc2ncnc3ccccc23)cc1'
+    sn = StateNetwork()
+    sn.build(smiles=smiles, max_formal_charge=3, tautomer_rule=None)
+    assert len(sn.visited_states) == 12
+    assert len(sn.graph.nodes()) == 12
+    se = sn.get_state_ensemble()
+    return se 
+
+
+def test_state_serialize(prepared_sm07_state):
+    st = prepared_sm07_state
     serialized = st.serialize()
     st2 = State().deserialize(serialized)
     assert st.smiles == st2.smiles
@@ -26,9 +37,9 @@ def test_state_serialize(prepared_state):
     assert st.charge == st2.charge
 
 
-def test_site(prepared_state):
+def test_site(prepared_sm07_state):
     """Ionizable site"""
-    st = prepared_state
+    st = prepared_sm07_state
     assert st.site_info() == [
         ('N', 5, 0, True, True),
         ('N', 7, 0, True, False), 
@@ -69,8 +80,8 @@ def test_tautomers():
     assert se2.size() == 20
 
 
-def test_protonate(prepared_state):
-    st = prepared_state
+def test_protonate(prepared_sm07_state):
+    st = prepared_sm07_state
     ps = st.get_protonated(atom_idx=9)
     assert len(ps) == 1
     assert ps[0].site_info() == [
@@ -105,8 +116,8 @@ def test_protonate(prepared_state):
     assert set(results) == set(expected)
 
 
-def test_deprotonate(prepared_state):
-    st = prepared_state
+def test_deprotonate(prepared_sm07_state):
+    st = prepared_sm07_state
     des = st.get_deprotonated(atom_idx=5)
     assert len(des) == 1
     assert des[0].site_info() == [
@@ -132,23 +143,7 @@ def test_deprotonate(prepared_state):
     expected = ['c1ccc(C[N-]c2ncnc3ccccc23)cc1']
     assert set(results) == set(expected)
 
-
-# def test_sn():
-#     smiles = 'c1ccc(CNc2ncnc3ccccc23)cc1'
-#     sn1 = StateNetwork()
-#     sn1.build(smiles=smiles, tautomer_rule=None)
-#     assert len(sn1.visited_states) == 11
-#     assert len(sn1.graph.nodes()) == 11
-#     sn2 = StateNetwork()
-#     sn2.build(smiles=smiles, tautomer_rule='rdkit')
-#     assert len(sn2.visited_states) == 33
-#     assert len(sn2.graph.nodes()) == 33
-#     sn3 = StateNetwork()
-#     sn3.build(smiles=smiles, tautomer_rule='comprehensive')
-#     assert len(sn3.visited_states) == 183
-#     assert len(sn3.graph.nodes()) == 183
     
-
 def test_unipka_workflow():
     smiles = 'c1ccc(CNc2ncnc3ccccc23)cc1'
     sn = StateNetwork()
@@ -171,17 +166,15 @@ def test_unipka_workflow():
     assert sn.get_num_nodes() == sn.get_state_ensemble().size()
 
     # calculated from Uni-pKa
-    LN10 = math.log(10)
-    TRANSLATE_PH = 6.504894871171601
     # Uni-pka model specific variable for pH dependent deltaG
     # Training might be conducted with a dataset in which raw pKa values
     # were subtracted by the mean value (TRANSLATE_PH).
-    FE = np.array([-6.025253772735596, -2.9201512336730957, -2.7405877113342285, 
+    dG = np.array([-6.025253772735596, -2.9201512336730957, -2.7405877113342285, 
           -2.9639060497283936, 7.656927108764648, 19.67357063293457, 
           21.269811630249023, 11.911577224731445, 7.5623698234558105, 
           10.144123077392578, 21.36874008178711, 12.132856369018555])
     
-    sn.set_energies(FE)
+    sn.set_energies(dG, ref_ph=6.504894871171601)
     se = sn.get_state_ensemble()
 
     se2 = StateEnsemble([st for st in sn.get_state_ensemble()])
@@ -197,32 +190,9 @@ def test_unipka_workflow():
     assert se == se3
     assert all(math.isclose(x, y) for x, y in zip([st.energy for st in se], [st.energy for st in se3]))
 
-    # micro-pKa
-    micro = se.get_micro_pKa(beta=1.0)
-    micro_pKa = {}
-    for k, v in micro.items():
-        # translation is necessary!!
-        micro_pKa[k] = (np.array(v)/LN10 + TRANSLATE_PH).tolist()
-    expected_micro_pKa = {
-        5: [5.263056256779453, 12.37074708270436], 
-        7: [5.190629233663213], 
-        9: [5.280577863433882],
-        }
-    assert micro_pKa.keys() == expected_micro_pKa.keys()
-    for k,v in micro_pKa.items():
-        for x, y in zip(v, expected_micro_pKa[k]):
-            assert math.isclose(x, y)
-
-    # macro-pKa
-    macro = se.get_macro_pKa(beta=1.0)
-    expected_macro_pKa = [5.248700101444452, 12.3693201962264]
-    macro_pKa = (np.array(macro)/LN10 + TRANSLATE_PH).tolist()
-    for x, y in zip(macro_pKa, expected_macro_pKa):
-        assert math.isclose(x, y)
-
     # population
-    pH = np.array([1.2, 7.4, 14.0]) - TRANSLATE_PH
-    p = sn.get_population(pH, C=LN10, beta=1.0)
+    ph_values = np.array([1.2, 7.4, 14.0])
+    p = sn.get_population(ph_values, C=math.log(10), beta=1.0)
     expected_p = [
         [2.6935020358203173e-05, 0.983819163640424, 0.027227135908720494], 
         [0.24360336373466238, 0.005614114302596941, 3.9027248338257366e-11], 
@@ -244,19 +214,13 @@ def test_unipka_workflow():
 
     print("\n")
     for k, st in enumerate(sn.visited_states):
-        print(f"{k:2} {st.smiles:50} {FE[k]:8.3f} {p[k]}")
+        print(f"{k:2} {st.smiles:50} {dG[k]:8.3f} {p[k]}")
 
     # population chart
     pH = np.linspace(0, 14, 60)
-    p = sn.get_population(pH=pH-TRANSLATE_PH, C=LN10, beta=1.0)
+    p = sn.get_population(ph_values=pH, C=math.log(10), beta=1.0)
 
-    plot_data, populated_state_idx = sn.get_plot_data_pH_vs_population(
-                                        pH=pH-TRANSLATE_PH, 
-                                        ignore_below=0.05,
-                                        pH_label=pH, 
-                                        C=LN10, 
-                                        beta=1.0,
-                                        )
+    plot_data, populated_state_idx = sn.get_plot_data_pH_vs_population(pH, p)
     
     print(f"Number of populated states: {len(populated_state_idx)}")
     print(f"Populated states (index in the state ensemble): {populated_state_idx}")
@@ -267,48 +231,159 @@ def test_unipka_workflow():
 
 
 def test_trim_and_sort():
-    TRANSLATE_PH = 6.504894871171601
-
     smiles = 'c1ccc(CNc2ncnc3ccccc23)cc1'
     sn = StateNetwork()
     sn.build(smiles=smiles, max_formal_charge=3, tautomer_rule=None)
-    FE = np.array([-6.025253772735596, -2.9201512336730957, -2.7405877113342285, 
+    dG = np.array([-6.025253772735596, -2.9201512336730957, -2.7405877113342285, 
           -2.9639060497283936, 7.656927108764648, 19.67357063293457, 
           21.269811630249023, 11.911577224731445, 7.5623698234558105, 
           10.144123077392578, 21.36874008178711, 12.132856369018555])
     
-    sn.set_energies(FE)
+    sn.set_energies(dG, ref_ph=6.504894871171601)
     se = sn.get_state_ensemble()
 
     assert sn.size() == 12
     assert se.size() == 12
 
-    pH = np.linspace(0.0, 14.0, 60) -  TRANSLATE_PH
+    ph_values = np.linspace(0.0, 14.0, 60)
+    p = sn.get_population(ph_values, C=math.log(10), beta=1.0)
     
-    sn.trim(pH, threshold=0.05)
+    sn.trim(p, threshold=0.05)
     assert sn.size() == 6
 
-    se.trim(pH, threshold=0.05)
+    se.trim(p, threshold=0.05)
     assert se.size() == 6
 
-    physiological_pH = np.array([7.4]) -  TRANSLATE_PH
+    physiological_pH = np.array([7.4])
 
-    p = se.get_population(physiological_pH)
-    assert np.allclose(p, np.array([[9.83820473e-01],
+    p74 = se.get_population(ph_values=physiological_pH)
+    assert np.allclose(p74, np.array([[9.83820473e-01],
                                    [5.61412177e-03], 
                                    [4.69135591e-03],
                                    [5.86521996e-03],
                                    [8.82928077e-06],
                                    [2.57359439e-15]]))
     
-
-    se.sort(physiological_pH)
+    se.sort(p74)
     assert se.size() == 6
-
-    p = se.get_population(physiological_pH)
-    assert np.allclose(p, np.array([[9.83820473e-01],
+    p74_ = se.get_population(ph_values=physiological_pH)
+    assert np.allclose(p74_, np.array([[9.83820473e-01],
                                     [5.86521996e-03],
                                     [5.61412177e-03],
                                     [4.69135591e-03],
                                     [8.82928077e-06],
-                                    [2.57359439e-15]]))
\ No newline at end of file
+                                    [2.57359439e-15]]))
+
+
+def test_macropka(prepared_sm07_state_ensemble):
+    se = prepared_sm07_state_ensemble
+
+    # calculated from Uni-pKa
+    # Uni-pka model specific variable for pH dependent deltaG
+    # Training might be conducted with a dataset in which raw pKa values
+    # were subtracted by the mean value (TRANSLATE_PH).
+    dG = np.array([
+        -6.025253772735596, 
+        -2.9201512336730957, 
+        -2.7405877113342285, 
+          -2.9639060497283936, 
+          7.656927108764648, 
+          19.67357063293457, 
+          21.269811630249023, 
+          11.911577224731445, 
+          7.5623698234558105, 
+          10.144123077392578, 
+          21.36874008178711, 
+          12.132856369018555])
+    
+    se.set_energies(dG, ref_ph=6.504894871171601)
+
+    assert se.get_charge_groups() == {-1: [4], 0: [0, 8, 9], 1: [1, 2, 3, 11], 2: [5, 6, 7], 3: [10]}
+
+    ph_values = np.linspace(-5, 20, 250)
+    p = se.get_population(ph_values, C=math.log(10), beta=1.0)
+
+    # population should be the same regardless of ref_state_idx
+    # p_1 = se.get_population(ph_values, C=LN10, beta=1.0, ref_state_idx=1)
+    # p_2 = se.get_population(ph_values, C=LN10, beta=1.0, ref_state_idx=2)
+    # for x, y, z in zip(p, p_1, p_2):
+    #     for xx, yy, zz in zip(x, y, z):
+    #         assert math.isclose(xx, yy, abs_tol=0.01)
+    #         assert math.isclose(xx, zz, abs_tol=0.01)
+
+    # calculated macro pKa should be within 1 pH unit.
+    macro_pKa = se.get_macro_pKa(ph_values, p)
+    expt_macro_pKa = 6.08
+    print(f"\nexpt={expt_macro_pKa} calc={macro_pKa}")
+    assert any([math.isclose(x, expt_macro_pKa, abs_tol=0.8) for x,y in macro_pKa])
+
+    # build state ensemble manually
+    # reference: 
+    # https://github.com/samplchallenges/SAMPL6/blob/master/physical_properties/pKa/microstates/SM07_microstates.csv
+    # microstate ID	canonical isomeric SMILES	canonical SMILES
+    # SM07_micro002	c1ccc(cc1)CN=c2c3ccccc3[nH]cn2	c1ccc(cc1)CN=c2c3ccccc3[nH]cn2
+    # SM07_micro003	c1ccc(cc1)C/N=c\2/c3ccccc3nc[nH]2	c1ccc(cc1)CN=c2c3ccccc3nc[nH]2
+    # SM07_micro004	c1ccc(cc1)CNc2c3ccccc3ncn2	c1ccc(cc1)CNc2c3ccccc3ncn2
+    # SM07_micro006	c1ccc(cc1)C/[NH+]=c/2\c3ccccc3[nH]cn2	c1ccc(cc1)C[NH+]=c2c3ccccc3[nH]cn2
+    # SM07_micro007	c1ccc(cc1)CNc2c3ccccc3nc[nH+]2	c1ccc(cc1)CNc2c3ccccc3nc[nH+]2
+    # SM07_micro011	c1ccc(cc1)C[NH2+]c2c3ccccc3ncn2	c1ccc(cc1)C[NH2+]c2c3ccccc3ncn2
+    # SM07_micro012	c1ccc(cc1)C[N-]c2c3ccccc3ncn2	c1ccc(cc1)C[N-]c2c3ccccc3ncn2
+    # SM07_micro013	c1ccc(cc1)C[NH2+]c2c3ccccc3[nH+]cn2	c1ccc(cc1)C[NH2+]c2c3ccccc3[nH+]cn2
+    # SM07_micro014	c1ccc(cc1)CNc2c3ccccc3[nH+]c[nH+]2	c1ccc(cc1)CNc2c3ccccc3[nH+]c[nH+]2
+    # SM07_micro015	c1ccc(cc1)C[NH2+]c2c3ccccc3nc[nH+]2	c1ccc(cc1)C[NH2+]c2c3ccccc3nc[nH+]2
+    # SM07_micro016	c1ccc(cc1)C[NH2+]c2c3ccccc3[nH+]c[nH+]2	c1ccc(cc1)C[NH2+]c2c3ccccc3[nH+]c[nH+]2
+    states = []
+    for smiles in [
+        'c1ccc(cc1)CN=c2c3ccccc3[nH]cn2',
+        'c1ccc(cc1)CN=c2c3ccccc3nc[nH]2',
+        'c1ccc(cc1)CNc2c3ccccc3ncn2',
+        'c1ccc(cc1)C[NH+]=c2c3ccccc3[nH]cn2',
+        'c1ccc(cc1)CNc2c3ccccc3nc[nH+]2',
+        'c1ccc(cc1)C[NH2+]c2c3ccccc3ncn2',
+        'c1ccc(cc1)C[N-]c2c3ccccc3ncn2',
+        'c1ccc(cc1)C[NH2+]c2c3ccccc3[nH+]cn2',
+        'c1ccc(cc1)CNc2c3ccccc3[nH+]c[nH+]2',
+        'c1ccc(cc1)C[NH2+]c2c3ccccc3nc[nH+]2',
+        'c1ccc(cc1)C[NH2+]c2c3ccccc3[nH+]c[nH+]2',
+        ]:
+        states.append(State(smiles=smiles))
+        
+    sem = StateEnsemble(states)
+    sem.info()
+
+
+
+
+def test_micropka(prepared_sm07_state_ensemble):
+    se = prepared_sm07_state_ensemble
+    
+    # calculated from Uni-pKa
+    # Uni-pka model specific variable for pH dependent deltaG
+    # Training might be conducted with a dataset in which raw pKa values
+    # were subtracted by the mean value (TRANSLATE_PH).
+    dG = np.array([
+        -6.025253772735596, 
+        -2.9201512336730957, 
+        -2.7405877113342285, 
+          -2.9639060497283936, 
+          7.656927108764648, 
+          19.67357063293457, 
+          21.269811630249023, 
+          11.911577224731445, 
+          7.5623698234558105, 
+          10.144123077392578, 
+          21.36874008178711, 
+          12.132856369018555])
+
+    se.set_energies(dG, ref_ph=6.504894871171601)
+    ph_values = np.linspace(-5, 20, 250)
+    p = se.get_population(ph_values, C=math.log(10), beta=1.0)
+
+    charge_groups_per_site = se.get_charge_groups_per_site()
+    assert charge_groups_per_site == {
+        5: {0: [0, 2, 3, 7], 1: [1, 5, 6, 10], -1: [4, 8, 9, 11]}, 
+        7: {0: [0, 1, 3, 4, 6, 9], 1: [2, 5, 7, 8, 10, 11]}, 
+        9: {0: [0, 1, 2, 4, 5, 8], 1: [3, 6, 7, 9, 10, 11]}}
+    
+    micropka = se.get_micro_pKa(ph_values, p)
+    print(micropka)
\ No newline at end of file
