@@ -181,4 +181,9 @@ def test_complete_tautomers():
     
     m = Mol("CCCCNC(=O)[C@@H]1CCCN1[C@@H](CC)c1nnc(Cc2ccc(C)cc2)o1")
     libr = rdworks.complete_tautomers(m)
-    assert libr.count() == 5
\ No newline at end of file
+    assert libr.count() == 5
+
+
+def test_optimize_confs():
+    mol = Mol("N=C1O[C@@H](CN1)CN2C[C@H](O[C@H](C2)C)C", "stereoisomer")
+    mol = mol.make_confs().compute(max_workers=4, chunksize=1, progress=True).optimize_confs(calculator='MMFF94')
\ No newline at end of file
